/* File: main.c
 * Desc: Slave handler allowing nodes to talk to CardServices
 * Auth: San Mehat
 * Note: 
 */      

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <signal.h>

#include <znodelist.h>
#include <afp.h>

extern int errno;

typedef struct configchain
	{
	AFP_CONFIG *Tconfigp_ptr;
	char       SZ_localmountpoint[255];
	int				 I_pid;                   /* PID of webserver when running */
	struct configchain *next;
	} CONFIGCHAIN;

CONFIGCHAIN *Tconfigchainp_anchor=NULL;

void handle_signal(int I_signo)
{
	switch (I_signo)
		{
		case SIGINT:
			unmount_filesystems();
			exit(-1);
			break;
		}
}

int main(argc,argv)
char **argv;
{
	int                I_sockfd;
	struct sockaddr_in outbound_addr;
	struct sockaddr_in inbound_addr;
	void               *Vp_inbuff;
	void               *Vp_outbuff;
	int                I_state=0;
	AFP_HDR            *in_hdr;
	AFP_HDR            *out_hdr;

	signal(SIGINT,handle_signal);

	system("clear");
	/* Initialize socket system */
	disp_label("Initializing socket system");	 
	if (!(I_sockfd=socket(AF_INET,SOCK_DGRAM,0))<0)
		{
		perror("socket");
		exit(1);
		}

	bzero((char *) &outbound_addr,sizeof(outbound_addr));
	outbound_addr.sin_family      = AF_INET;
//	outbound_addr.sin_addr.s_addr = inet_addr("10.1.54.6");
	outbound_addr.sin_addr.s_addr = inet_addr("192.168.77.1");
	outbound_addr.sin_port        = htons(6442);

	bzero((char *) &inbound_addr,sizeof(inbound_addr));
	inbound_addr.sin_family      = AF_INET;
	inbound_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	inbound_addr.sin_port        = htons(6443);

  if (bind(I_sockfd, (struct sockaddr *) &inbound_addr,sizeof(inbound_addr))<0)
    {
    perror("bind");
    exit(1);
    }                        

	if (!(Vp_inbuff=(void *)malloc(MAX_AFPSIZE)))
		{
		disp_label("Unable to allocate memory for asynch input buffer");
		exit(1);
		}
	printf("[%s] Allocated %d bytes at 0x%x for asynch input buffer\n",mytime(),MAX_AFPSIZE,Vp_inbuff);
	if (!(Vp_outbuff=(void *)malloc(MAX_AFPSIZE)))
		{
		disp_label("Unable to allocate memory for asynch output buffer");
		exit(1);
		}
	printf("[%s] Allocated %d bytes at 0x%x for asynch output buffer\n",mytime(),MAX_AFPSIZE,Vp_outbuff);
	in_hdr= (AFP_HDR *)Vp_inbuff;
	out_hdr= (AFP_HDR *)Vp_outbuff;
#define STATE_INIT       0
#define STATE_SEND       1
#define STATE_WAIT       2
#define STATE_PROCESS    3
#define STATE_EXIT       4
#define STATE_READY      5
	while(1)
		{
		int I_rc;

		if (I_state==STATE_INIT)
			{
			//sleep(1);
			out_hdr->UI_txid=random();
			out_hdr->UI_cmd=REQ_CONFIG;
			out_hdr->UI_dlen=0;
			I_state=STATE_SEND;
			}
		else if (I_state==STATE_SEND)
			{
#if 0
			printf("[%s] Sending packet of size %d bytes (0x%x)\n",
				mytime(),
				(sizeof(AFP_HDR)+out_hdr->UI_dlen),(sizeof(AFP_HDR)+out_hdr->UI_dlen));
#endif
			if ((I_rc=sendto(I_sockfd,Vp_outbuff,(sizeof(AFP_HDR)+out_hdr->UI_dlen),0,(struct sockaddr *)&outbound_addr,sizeof(outbound_addr)))<0)
				{
				perror("sendto");
				exit(1);
				}
			I_state=STATE_WAIT;
			}
		else if (I_state==STATE_WAIT)
			{
			struct timeval to;
			int            I_rc;
			fd_set         readfds;

			to.tv_sec=10;
			to.tv_usec=0;
			FD_ZERO(&readfds);
			FD_SET(I_sockfd,&readfds);
//			disp_label("Waiting for packet");
			if ((I_rc=select((I_sockfd+1),&readfds,0,0,&to))<0)
				{
				perror("select");
				exit(1);
				}
			else if (!I_rc)
				{
				if (!Tconfigchainp_anchor)
					{
					/* We have timed out but as yet have not received any configuration data
             Uh oh.... resend configuration request
           */
					I_state=STATE_INIT;
					continue;
					}
				disp_label("Performing periodic processing of event queue");
				process_outstanding_event_queue();
				continue;
				}
			if (FD_ISSET(I_sockfd,&readfds))
				{
				int     len=sizeof(inbound_addr);

				if ((I_rc=recvfrom(I_sockfd,Vp_inbuff,MAX_AFPSIZE,0,(struct sockaddr *)&inbound_addr,&len))<0)
					{
					perror("recvfrom");
					exit(1);
					}
				if (!I_rc)
					{
					disp_label("Ignoring rc of 0 from read()");
					continue;
					}
// printf("[%s] Received %d bytes from controller (txid=%d)\n",mytime(),I_rc,in_hdr->UI_txid);
				I_state=STATE_PROCESS;
				}
			else
				{
				printf("[CardModule] select() assert! failure\n");
				exit(1);
				}
			I_state=STATE_PROCESS;
			}
		else if (I_state==STATE_PROCESS)
			{
			switch (in_hdr->UI_cmd)
				{
				case REQ_RESET:
					{
					printf("[%s] Controller performing system reset\n",mytime());
					stop_webservers();
					unmount_filesystems();
					I_state=STATE_INIT;
					break;
					}
				case REQ_STATUS:
					{
					Z_STATS *statsptr;
					char    *tmpbuff=Vp_outbuff;

					printf("[%s] Controller STATUS query received\n",mytime());
					memset(Vp_outbuff,0,(sizeof(AFP_HDR)+sizeof(Z_STATS)));
					statsptr = (Z_STATS*) &tmpbuff[sizeof(AFP_HDR)];
					/* Obtain system statistics and crap */
					get_statistics(statsptr);
					/* Now send the data to the controller */
					out_hdr->UI_txid=in_hdr->UI_txid;
					out_hdr->UI_cmd =ACK_STATUS;
					out_hdr->UI_dlen = sizeof(Z_STATS);
					I_state=STATE_SEND;
					break;
					}
				case ACK_CONFIG:
					{
					AFP_CONFIG *Tconfigp_ptr;
					int         I_numpartitions;
					int         I_c1;
					char        *buff= (char *)Vp_inbuff;

					printf("[%s] Configuration ACK / DATA packet received\n",mytime());
					//sleep(1);

					/* First we want to tear down our old partition list */
					{
					CONFIGCHAIN *Tconfigchainp_walk;
				
					Tconfigchainp_walk=Tconfigchainp_anchor;
					while(Tconfigchainp_walk)
						{
						CONFIGCHAIN *Tconfigchainp_erase=Tconfigchainp_walk;
						AFP_CONFIG  *Tafpconfigp_erase=Tconfigchainp_walk->Tconfigp_ptr;
						
						Tconfigchainp_walk=Tconfigchainp_walk->next;
						free(Tafpconfigp_erase);
						free(Tconfigchainp_erase);
						}
					Tconfigchainp_anchor=NULL;
					}

					I_numpartitions=(in_hdr->UI_dlen/sizeof(AFP_CONFIG));
					printf("[%s] Detected %d partitions for this node\n",mytime(),I_numpartitions);

					// Create the partition configuration chain and link it all up
					for (I_c1=0,
							 Tconfigp_ptr=(AFP_CONFIG *) &buff[sizeof(AFP_HDR)];
							 I_c1<I_numpartitions;
							 I_c1++,Tconfigp_ptr++)
						{
						CONFIGCHAIN *Tconfigchainp_new;

						Tconfigchainp_new= (CONFIGCHAIN *) malloc (sizeof(CONFIGCHAIN));
						memset(Tconfigchainp_new,0,sizeof(CONFIGCHAIN));
						Tconfigchainp_new->Tconfigp_ptr= (AFP_CONFIG *) malloc (sizeof(AFP_CONFIG));
						memset(Tconfigchainp_new->Tconfigp_ptr,0,sizeof(AFP_CONFIG));
						if (!Tconfigchainp_anchor)
							Tconfigchainp_anchor=Tconfigchainp_new;
						else
							{
							CONFIGCHAIN *Tconfigchainp_walk;
							
							Tconfigchainp_walk=Tconfigchainp_anchor;
							while (Tconfigchainp_walk->next)
								Tconfigchainp_walk=Tconfigchainp_walk->next;
							Tconfigchainp_walk->next = Tconfigchainp_new;
							}
						strcpy(Tconfigchainp_new->Tconfigp_ptr->SZ_docroot,Tconfigp_ptr->SZ_docroot);
						Tconfigchainp_new->Tconfigp_ptr->US_portbinding=Tconfigp_ptr->US_portbinding;
						sprintf(Tconfigchainp_new->SZ_localmountpoint,"/mpoints/partition-%d",I_c1);
						printf("[%s] (%d)[0x%x] PB=%d DR=%s\n",
							mytime(),
							I_c1,
							Tconfigchainp_new,
							Tconfigchainp_new->Tconfigp_ptr->US_portbinding,
							Tconfigchainp_new->Tconfigp_ptr->SZ_docroot);
						}
					/* 
						 XXX: Do nfs mounts here, start webservers, passing port binding and docroot on 
						 cmdline 
					 */
					if (mount_filesystems()<0)
						{
						printf("[%s] Unable to mount all filesystems\n",mytime());
						I_state=STATE_EXIT;
						}
					else
						{
						if (start_webservers()<0)
							{
							printf("[%s] Unable to start all webservers\n",mytime());
							I_state=STATE_EXIT;
							}
						else
							{
							printf("[%s] Webservers started\n",mytime());
							I_state=STATE_READY;
							}
						}
					break;
					}
				default:
					{
					printf("[%s] Node sent invalid command (%d)\n",mytime(),in_hdr->UI_cmd);
					I_state=STATE_WAIT;
					break;
					}
				}

			process_outstanding_event_queue();
			}
		else if (I_state==STATE_EXIT)
			{
			disp_label("Normal shutdown proceedure");
			free(Vp_inbuff);
			free(Vp_outbuff);
			close(I_sockfd);
			unmount_filesystems();
			exit(0);
			}
		else if (I_state==STATE_READY)
			{
			disp_label("Building ready signal");
			memset(out_hdr,0,sizeof(AFP_HDR));
			out_hdr->UI_txid=random();
			out_hdr->UI_cmd=SIG_READY;
			out_hdr->UI_dlen=0;
			I_state=STATE_SEND;
		//	sleep(4);
			}
		}
}

process_outstanding_event_queue()
{

}

disp_label(char *msg)
{
	printf("[%s] %s\n",mytime(),msg);
}

/* Debug thing for timestamps */
char timebuff[64];
char *mytime()
{
  time_t timeval;

  timeval=time(NULL);
  strcpy(timebuff,ctime(&timeval));
  timebuff[(strlen(timebuff)-1)]=0;
  return(timebuff);
}
                         
get_statistics(Z_STATS *statsptr)
{
	FILE *Tfilep_ptr;
	char SZ_line[255];
	int  holder;
	float F_tmp;

  /* Do cpuload */
	if (!(Tfilep_ptr=fopen("/proc/loadavg","r")))
		{
		printf("Unable to get system load average");
		return(-1);
		}
	fscanf(Tfilep_ptr,"%f",&F_tmp);
	statsptr->UI_cpuload= (unsigned int) (F_tmp*100);
	fclose(Tfilep_ptr);

	/* Do temperature */

	/* Do memory */
	if (!(Tfilep_ptr=fopen("/proc/meminfo","r")))
		{
		printf("Unable to get system load average");
		return(-1);
		}
	fgets(SZ_line,sizeof(SZ_line),Tfilep_ptr);
	fscanf(Tfilep_ptr,"%s %d %d %d",SZ_line,&statsptr->UI_memtotal,&holder,&statsptr->UI_memfree);
	fclose(Tfilep_ptr);
}

int mount_filesystems()
{
	int I_rc;
	char SZ_fstype[10];
	CONFIGCHAIN *Tconfigchainp_ptr;
	char SZ_tmp[255];


	for (Tconfigchainp_ptr=Tconfigchainp_anchor;
			Tconfigchainp_ptr;
			Tconfigchainp_ptr=Tconfigchainp_ptr->next)
		{
		/* Check to see if the mountpoint exists */
		if (access(
				Tconfigchainp_ptr->SZ_localmountpoint,
				R_OK|W_OK|X_OK)<0)
			{
			printf("[%s] Unable to access %s\n",
				mytime(),
				Tconfigchainp_ptr->SZ_localmountpoint);
			sprintf(SZ_tmp,"/bin/mkdir -p %s",Tconfigchainp_ptr->SZ_localmountpoint);	
			system(SZ_tmp);
			printf("[%s] Directory created\n",mytime());
			}
		printf("[%s] Attempting to mount %s\n",
			mytime(),
			Tconfigchainp_ptr->Tconfigp_ptr->SZ_docroot);
		sprintf(SZ_tmp,"/bin/mount %s %s >&/dev/null",
			Tconfigchainp_ptr->Tconfigp_ptr->SZ_docroot,
			Tconfigchainp_ptr->SZ_localmountpoint);
		I_rc=system(SZ_tmp);
		I_rc=I_rc>>8;
		if (I_rc!=0)
			{
			printf("[%s] Filesystem mount failed(err=%d)\n",mytime(),I_rc);
			return(-1);
			}
		}
	return(0);
}

int unmount_filesystems()
{
	int I_rc;
	char SZ_fstype[10];
	CONFIGCHAIN *Tconfigchainp_ptr;
	char SZ_tmp[255];


	for (Tconfigchainp_ptr=Tconfigchainp_anchor;
			Tconfigchainp_ptr;
			Tconfigchainp_ptr=Tconfigchainp_ptr->next)
		{
		printf("[%s] Attempting to unmount %s\n",
			mytime(),
			Tconfigchainp_ptr->Tconfigp_ptr->SZ_docroot);
		sprintf(SZ_tmp,"/bin/umount %s >& /dev/null",
			Tconfigchainp_ptr->SZ_localmountpoint);
		I_rc=system(SZ_tmp);
		I_rc=I_rc>>8;
		if (I_rc!=0)
			{
			printf("[%s] Filesystem unmount failed (err=%d)\n",mytime(),I_rc);
			return(-1);
			}
		}
	return(0);
}

int start_webservers()
{
	CONFIGCHAIN *Tconfigchainp_ptr;

	for (Tconfigchainp_ptr=Tconfigchainp_anchor;
			Tconfigchainp_ptr;
			Tconfigchainp_ptr=Tconfigchainp_ptr->next)
		{
		/* XXX: Do the fork / exec here and record the PID */
		printf("[%s] Starting webserver on port %d with docroot %s\n",
			mytime(),
			Tconfigchainp_ptr->Tconfigp_ptr->US_portbinding,
			Tconfigchainp_ptr->Tconfigp_ptr->SZ_docroot);
		}
	return(0);
}

int stop_webservers()
{
	CONFIGCHAIN *Tconfigchainp_ptr;

	for (Tconfigchainp_ptr=Tconfigchainp_anchor;
			Tconfigchainp_ptr;
			Tconfigchainp_ptr=Tconfigchainp_ptr->next)
		{
		/* XXX: Do the kill here */
		printf("[%s] Sending SIGKILL to webserver on port %d with docroot %s (pid %d)\n",
			mytime(),
			Tconfigchainp_ptr->Tconfigp_ptr->US_portbinding,
			Tconfigchainp_ptr->Tconfigp_ptr->SZ_docroot,
			Tconfigchainp_ptr->I_pid);
		}
	return(0);


}
