/* 
 * ZAPHODCFG -- Zaphod Serial Configuration Program
 *
 * Copyright (C) 1999 by Zac Sprackett, <zacs@corelcomputer.com>
 * Based on NWCONFIG (C) 1998 by Ralph F. Siemsen, <ralphs@corelcomputer.com>
 * Revised and completed by Mark Knox, Feb/Mar 1999
 *
 */

#include <ncursui.h>  // User Interface Tools
#include "language"   // Language Bindings
#include <fcntl.h>
#include <config.h>   // Config Types
#include <netinet/in.h>
#include <signal.h>

#include <syslog.h>

// Global Variables
Window *Root;
Form *Top, *Sec;
CONFIGURATION *Tconfigp_ptr;	

// Function Prototypes
int ConfirmExit();
bool ConfirmDelete(char* SZ_server);
int menu_network(void);
int menu_misc();
int menu_virtuals();
int menu_choose_virt(int);
int menu_edit_virt(int);

enum
{
	STATE_NETWORK, STATE_VIRTUALS, /*STATE_PRIORITIES,*/
	STATE_MISCELLANEOUS, STATE_EXIT
};

// The following constants are used to construct the two menu items allowing
// the user to return to the main menu or add a new virtual server.
// These occur in the virtual server menu list.
#define NEW_VIRTUAL_SERVER "<<New Virtual Server>>"



int main( int argc, char **argv )
{
	int I_State;

	// Check that root is running this
	if (getuid() != 0)
	{
		printf ("Sorry, this program must be run by the root user only.\n");
		return 1;
	}
	
	openlog("zaphodcfg", 0, LOG_USER);

	// Menu Structures
	struct MenuItem mi1 = { NULL, 'N', "etwork", 0, 0 };
	struct MenuItem mi2 = { NULL, 'W', "eb Servers", 0, 0 };
//	struct MenuItem mi3 = { NULL, 'A', "ssign node priority", 0, 0 };
	struct MenuItem mi4 = { NULL, 'M', "iscellaneous", 0, 0 };
	struct MenuItem mi5 = { "E", 'x', "it the configuration utility", 0, 0 };
	struct MenuItem *mia[] = { &mi1, &mi2, /*&mi3,*/ &mi4, &mi5 };
	struct Menu MainMenu =
	{
 		4, 1, 4, mia
	};

	// Load our config file
	if (!(Tconfigp_ptr=(CONFIGURATION *) malloc(sizeof(CONFIGURATION))))
	{
		printf( "Unable to allocate memory for configuration data!\n" );
		exit (1);
	}
	bzero( Tconfigp_ptr, sizeof(CONFIGURATION));
	if (config_load(Tconfigp_ptr)<0)
	{
		printf( "Unable to load configuration data!\n" );
		exit (1);
	}
	
	// Convert the config structure into our Stored array *sigh*
	serverToHostAndDomain(	Tconfigp_ptr->SZ_servername,
				Stored[HOSTNAME],
				Stored[DOMAIN]);
	sprintf(Stored[EXTERNAL_IPADDR],"%s",itod(ntohl(Tconfigp_ptr->UI_ext_ipaddr)));
	sprintf(Stored[EXTERNAL_NETMASK],"%s",itod(ntohl(Tconfigp_ptr->UI_ext_netmask)));
	sprintf(Stored[INTERNAL_IPADDR],"%s",itod(ntohl(Tconfigp_ptr->UI_int_ipaddr)));
	sprintf(Stored[INTERNAL_NETMASK],"%s",itod(ntohl(Tconfigp_ptr->UI_int_netmask)));
	sprintf(Stored[GATEWAY],"%s",itod(ntohl(Tconfigp_ptr->UI_gateway)));
	sprintf(Stored[NAMESERVER],"%s",itod(ntohl(Tconfigp_ptr->UI_nameserver)));
	sprintf(Stored[NFSSERV],"%s",itod(ntohl(Tconfigp_ptr->UI_nfsserver)));
	sprintf(Stored[NFSBIN],"%s",Tconfigp_ptr->SZ_binarypath);
	sprintf(Stored[TIMEZONE],"%s",Tconfigp_ptr->SZ_timezone);
	sprintf(Stored[SERVERADMIN],"%s",Tconfigp_ptr->SZ_email);
	sprintf(Stored[INTERVAL],"%d seconds",Tconfigp_ptr->UI_cardsvcsscanfreq);
	sprintf(Stored[LOADBAL], "%s",Tconfigp_ptr->SZ_globalloadbalancer);

	// Main Window Setup
	Window RootWin (0,0,0,0);
	RootWin.Title (ProgHead());
	RootWin.Help (MenuHelp);
	RootWin.Update();
	Root = &RootWin;

	// Display our splash screen
	{
		Window Splash (5,6,70,14);
		Splash.Border (1);
		Splash.Update ();
		Window Inner (8,8,65,10);
		Inner.Print (0,0,SplashMsg); 
		Inner.Update();
		while (getch() == ERR);
	}

	RootWin.Update();

	// Main Loop
	while ( I_State != STATE_EXIT )
	{
		// Main Configuration Menu
		Window Main ( 2,2,76,14 );
		Main.Border (1);
		Main.Title (WinTitle[0]);
		Window MainBox ( 4, 4, 72, 10 );
		RootWin.Help(MenuHelp);
	
		Window Help ( 2,16,76,6);
		Help.Border (1);
		Help.Title (HelpTitle[0]);
		Window HelpBox ( 4,18,72,3 );

		Main.Update();
		Help.Update();
		I_State = MainBox.Select( &MainMenu, 0);

		// Hide the main panels
		hide_panel (MainBox.pan);
		hide_panel (Main.pan);
		hide_panel (Help.pan);
		RootWin.Update();

		switch (I_State)
		{
			case STATE_NETWORK:
				menu_network();
				break;

			case STATE_VIRTUALS:
				menu_virtuals();
				break;
/*
			case STATE_PRIORITIES:
				menu_priorities();
				break;
*/
			case STATE_MISCELLANEOUS:
				menu_misc();
				break;
		}		
	}

	// Confirm our quit
	ConfirmExit();

	closelog();
}

int menu_network(void)
{
	int ch;
	
	Root->Help(EditHelp);
	
	Window Main ( 2,2,76,14 );
	Main.Border (1);
	Main.Title (WinTitle[5]);
	Window MainBox ( 4, 3, 72, 12 );

	Window Help ( 2,16,76,6);
	Help.Border (1);
	Help.Title (HelpTitle[0]);
	Window HelpBox ( 4,17,72,3 );

	HostNameField host(&MainBox,1,26,40,HOSTNAME);
	TextField domain(&MainBox,2,26,40,DOMAIN);

	IPField ext_ipaddr(&MainBox,4,26,40,EXTERNAL_IPADDR);
	IPField ext_netmask(&MainBox,5,26,40,EXTERNAL_NETMASK);

	IPField int_ipaddr(&MainBox,7,26,40,INTERNAL_IPADDR);
	IPField int_netmask(&MainBox,8,26,40,INTERNAL_NETMASK);

	IPField gateway(&MainBox,10,26,40,GATEWAY);
	IPField nameserver(&MainBox,11,26,40,NAMESERVER);

//	SelectField zone(&MainBox,7,26,40,ZoneOptions,TIMEZONE);
//	TimeField t(&MainBox,8,26,TIMEDATE);

	Field *MainFormItems[] =
	{
		&host, &domain, &ext_ipaddr, &ext_netmask, &int_ipaddr,
		&int_netmask, &gateway, &nameserver, NULL
	};
	Form MainForm(MainFormItems, &HelpBox);
	Top = &MainForm;
	MainForm.Render();

	do {
		ch = getch();
		/*if(!t.changed) {
			Field *cur = MainForm.items[MainForm.selected];
			if(cur != &t) {
				t.Display(0);
				MainBox.Update();
			}
			cur->Display(!cur->changed);
			cur->win->Update();
		}*/
		ch = MainForm.Edit(ch);
	} while(ch != KEY_F(10));
	return 0;
}

int menu_virtuals(void)
{
	int index = -1;
	while ((index = menu_choose_virt(index)) != -1);
}

/*
 * Allows the user to create a new virtual server + scroll through the
 * existing virtual servers.
 *
 * index - the index of the virtual server to display initially, -1 =>
 *				select the default ie "new virtual server" item.
 *
 * Return -1 to return to the main menu or the index of the chosen virtual
 * server.
 */
int menu_choose_virt(int index)
{
	int			ch, I_Count, I_temp;
	CONFIG_PARTITION	*Tpartitionp_ptr, *prev;

	Window Main ( 2,2,76,14 );
	Main.Border (1);
	Main.Title (WinTitle[8]);
	Window MainBox ( 4, 3, 72, 10 );

	Window Help ( 2,16,76,6);
	Help.Border (1);
	Help.Title (HelpTitle[0]);
	Window HelpBox ( 4,17,72,3 );

	Root->Help(GenericHelp);

	// Allocate and build the list of hostnames
	char **hostnames;
	for (	I_Count = 0, Tpartitionp_ptr = Tconfigp_ptr->Tpartition_root;
		Tpartitionp_ptr;
        	I_Count++, Tpartitionp_ptr = Tpartitionp_ptr->next );

	// Account for the "new virtual server" item at the head of the list	
	if (I_Count > 0)
		index++;
	else
		index = 0;

	// Allocate enough for the "new virtual server" item and a NULL string
	hostnames = (char**)calloc(I_Count + 2, sizeof(char*));
	bzero(hostnames, (I_Count + 2) * sizeof(char*));
	hostnames[0] = NEW_VIRTUAL_SERVER;

	for (	I_temp = 1, Tpartitionp_ptr = Tconfigp_ptr->Tpartition_root;
        	Tpartitionp_ptr;
        	I_temp++, Tpartitionp_ptr = Tpartitionp_ptr->next )
	{
  		hostnames[I_temp] = Tpartitionp_ptr->SZ_servername;
	}
	
	// Store the first item in the list so that if no selection is made
	// and the user selects enter, it will be clear what menu item was
	// active at the time of selection
	strcpy(Stored[SERVERNAME],hostnames[index]);

	SelectField servers(&MainBox,2,16,40,hostnames, SERVERNAME);

	Field *MainFormItems[] =
	{
 		&servers, NULL
	};

	Form MainForm(MainFormItems, &HelpBox);
	Top = &MainForm;
	MainForm.Render();

	while(1)
	{
		ch = getch();
		MainForm.Edit(ch);
		servers.Validate();
		if ((ch == KEY_BACKSPACE) || (ch == KEY_DL)) //delete
		{
			// Remember the first item in the hostnames list allows addition of a
			// new virtual server and deletion on this item is not possible
			if (strcmp(Stored[SERVERNAME], hostnames[0]) != 0)
			{
				// Locate the item to delete in the partition list
				for(	I_temp = 0, prev = NULL,
							Tpartitionp_ptr = Tconfigp_ptr->Tpartition_root;
					strcmp(Tpartitionp_ptr->SZ_servername,Stored[SERVERNAME]) != 0;
					I_temp++, prev = Tpartitionp_ptr,
					Tpartitionp_ptr = Tpartitionp_ptr->next );

				// prompt for the user's approval
				if (ConfirmDelete(Tpartitionp_ptr->SZ_servername))
				{
					// Remember that the first one allows the user to add a new virtual
					// server so it cannot be deleted
					if (prev == NULL)
					{
						Tconfigp_ptr->Tpartition_root = Tpartitionp_ptr->next;
					}
					else
					{
						prev->next = Tpartitionp_ptr->next;
					}
					free(Tpartitionp_ptr);

					// Determine which item should be selected after this deletion.
					// Ensure that we don't overrun the list.
					I_Count--;
					if (I_temp < I_Count)
						index = I_temp;
					else if (I_Count > 0)
						index = I_Count - 1;
					else
						index = 0;
				}
			}
			else
			{
				index = 0;
			}
			free(hostnames);
			return index;	// return the index of the item adjacent to the deleted
		}
		else if (ch == KEY_IL)	//insert
		{
			free(hostnames);
			return menu_edit_virt(-1);
		}
		else if ((ch == KEY_ENTER) || (ch == 0x0A))	// edit
		{
			for (int I_temp = 0; hostnames[I_temp] != NULL; I_temp++)
			{
				if (strcmp(Stored[SERVERNAME], hostnames[I_temp]) == 0)
				{
					free(hostnames);
					return menu_edit_virt(I_temp - 1);
				}
			}
		}
		else if (ch == KEY_F(10))
		{
			free(hostnames);
			return -1;
		}
	}
}

/* 
 * Allows the user to edit or add a virtual server
 *
 * I_VSIndex - 0-based index of the item to edit, -1 to create a new item
 *
 * Return the 0-based index of the added/edited virtual server or -1 if the user
 * pressed F10.
 */
int menu_edit_virt(int I_VSIndex)
{
	int I_cnt, ch;
	CONFIG_PARTITION *Tpartition_ptr;
	CONFIG_NODELIST *node_ptr, *node_next;
	char buf[128];

	// clear all fields
	bzero(&Stored[V_HOSTNAME], (ITEM_MAX-V_HOSTNAME)*FIELD_MAX*sizeof(char));

	// if the index is -1 then we are adding a new virtual server
	// otherwise we are editing an existing one in which case the fields
	// need to be filled out with the appropriate contents
	if (I_VSIndex == -1)
	{
		strcpy(Stored[V_PORT],"80");	// default port number
	}
	else
	{
		for (	I_cnt = 0, Tpartition_ptr = Tconfigp_ptr->Tpartition_root;
			I_cnt < I_VSIndex;
			I_cnt++, Tpartition_ptr = Tpartition_ptr->next);

		serverToHostAndDomain(	Tpartition_ptr->SZ_servername,
					Stored[V_HOSTNAME],
					Stored[V_DOMAIN]);
		strcpy(Stored[V_IPADDR],itod(ntohl(Tpartition_ptr->UI_aliasaddr)));
		strcpy(Stored[V_DOCROOT],Tpartition_ptr->SZ_documentpath);
		strcpy(Stored[V_CGIROOT],Tpartition_ptr->SZ_cgipath);
		strcpy(Stored[V_LOADBAL],Tpartition_ptr->SZ_loadbalancer);
		strcpy(Stored[V_SSL],SSLOptions[Tpartition_ptr->C_ssl]);
		strcpy(Stored[V_SSLCERTIFICATE],Tpartition_ptr->SZ_sslcertificate);
		sprintf(Stored[V_PORT],"%d",ntohs(Tpartition_ptr->US_portbinding));
		switch(Tpartition_ptr->C_optionflags)
		{
			case OPT_INCLUDE_NOEXEC:
				strcpy(Stored[V_INCLUDES],IncludeOptions[1]);
				break;
			case OPT_INCLUDE_FULL:
				strcpy(Stored[V_INCLUDES],IncludeOptions[2]);
				break;
			default:
				strcpy(Stored[V_INCLUDES],IncludeOptions[0]);
		}
		I_cnt=0;
		memset(Stored[V_NODELIST],0,sizeof(Stored[V_NODELIST]));
		for (	node_ptr=Tpartition_ptr->Tnodelist_root;
			node_ptr;
			node_ptr = node_next)
		{
			node_next = node_ptr->next;
			Stored[V_NODELIST][I_cnt++] = node_ptr->I_nodenumber + '0';
			Stored[V_NODELIST][I_cnt++] = ' ';
			free(node_ptr);
		}
	}
	
	Root->Help(GenericHelp);
	
	Window Main ( 2,2,76,14 );
	Main.Border (1);
	Main.Title ((I_VSIndex == -1 ) ? WinTitle[6] : WinTitle[7]);
	Window MainBox ( 4,3,72,12 );

	Window Help ( 2,16,76,6);
	Help.Border (1);
	Help.Title (HelpTitle[0]);
	Window HelpBox ( 4,17,72,3 );

	// default load balancer
	HostNameField host(&MainBox,1,26,40,V_HOSTNAME);
	TextField domain(&MainBox,2,26,40,V_DOMAIN);
//	IPField ipaddr(&MainBox,3,26,40,V_IPADDR);
//	NumberField port(&MainBox,4,26,40,V_PORT);
	PathField docroot(&MainBox,4,26,40,V_DOCROOT);
	PathField cgiroot(&MainBox,5,26,40,V_CGIROOT);
	cgiroot.SetBlankAllowed(true);
	SelectField includes(&MainBox,6,26,40,IncludeOptions,V_INCLUDES);
//	SelectField loadbal(&MainBox,9,26,40,LoadBalancerOptions,V_LOADBAL);
	SelectField ssl(&MainBox,8,26,40,SSLOptions,V_SSL);
	PathField sslcertificate(&MainBox,9,26,40,V_SSLCERTIFICATE);
	sslcertificate.SetBlankAllowed(true);
	sslcertificate.SetDir(0);

	// We won't support assigning a list of nodes to a virtual server
	// in the first release so the following line of code is commented out
	// NumberListField nodes(&MainBox,9,16,24,V_NODELIST);
	Field *MainFormItems[] = {
		&host, &domain, /*&ipaddr, &port,*/ &docroot, &cgiroot,
		&includes, /*&loadbal, &nodes,*/ &ssl, &sslcertificate, NULL
	};
	Form MainForm(MainFormItems, &HelpBox);
	Top = &MainForm;
	MainForm.Render();

	do {
		ch = MainForm.Edit(getch());
	} while(ch != KEY_F(10));

	if(!strcmp(Stored[V_HOSTNAME],"") || !strcmp(Stored[V_DOMAIN],"")
	/*|| !strcmp(Stored[V_IPADDR],"")*/ || !strcmp(Stored[V_DOCROOT],"")
	/*|| !strcmp(Stored[V_CGIROOT],"") || !strcmp(Stored[V_PORT],"")*/)
		return 0; // FIXME: should have an error screen here

syslog( LOG_USER,
	"menu_edit_virt -> %sing %s.%s,%s:%s,doc=%s,cgi=%s,incl=%s,loadbal=%s",
	(I_VSIndex == -1) ? "Add" : "Edit",
        Stored[V_HOSTNAME],Stored[V_DOMAIN],Stored[V_IPADDR],Stored[V_PORT],
        Stored[V_DOCROOT],
        Stored[V_CGIROOT],
	Stored[V_INCLUDES],
	Stored[V_LOADBAL]);

	// If this was an addition, add a partition struct to the list
	if (I_VSIndex == -1)
	{
		CONFIG_PARTITION *ptmp;
		Tpartition_ptr = (CONFIG_PARTITION *)malloc(sizeof(CONFIG_PARTITION));
		Tpartition_ptr->next = Tconfigp_ptr->Tpartition_root;
		Tconfigp_ptr->Tpartition_root = Tpartition_ptr;
		I_VSIndex = 0;	// the index of the added partition
	}

	sprintf(Tpartition_ptr->SZ_servername, "%s.%s",
		Stored[V_HOSTNAME], Stored[V_DOMAIN]);
	Tpartition_ptr->UI_aliasaddr = htonl(dtoi(Stored[V_IPADDR]));
	Tpartition_ptr->UI_aliasmask = 0xffffffff;
	Tpartition_ptr->US_portbinding = htons((unsigned short)atoi(Stored[V_PORT]));
	strcpy(Tpartition_ptr->SZ_documentpath,Stored[V_DOCROOT]);
	strcpy(Tpartition_ptr->SZ_cgipath,Stored[V_CGIROOT]);

	// NOTE: Apr 7,1999 Since we only support name-based virtual hosting
	// load-balancing on a per virtual host basis is not supported. The
	// For field has been left in the structure to allow forward
	// compatibility when we eventually support IP-based virtual
	// hosting.
//	strcpy(Tpartition_ptr->SZ_loadbalancer,Stored[V_LOADBAL]);
	strcpy(	Tpartition_ptr->SZ_loadbalancer,
		Tconfigp_ptr->SZ_globalloadbalancer);
	
	// we don't use the route entry, so ignore it
	if(!strcmp(Stored[V_INCLUDES],"No Exec"))
		Tpartition_ptr->C_optionflags = OPT_INCLUDE_NOEXEC;
	else if(!strcmp(Stored[V_INCLUDES],"Full"))
		Tpartition_ptr->C_optionflags = OPT_INCLUDE_FULL;
	else
		Tpartition_ptr->C_optionflags = OPT_INCLUDE_NONE;
	
	Tpartition_ptr->Tnodelist_root = NULL;

	I_cnt = 0;
	while((ch=Stored[V_NODELIST][I_cnt++])!=0)
	{
		if(ch == ' ' || ch == ',')
			continue;
		node_ptr = (CONFIG_NODELIST *)malloc(sizeof(CONFIG_NODELIST));
		node_ptr->I_nodenumber = ch - '0';
		node_ptr->UI_throughputchoke = 0;
		node_ptr->UC_priority = 1;
		node_ptr->next = Tpartition_ptr->Tnodelist_root;
		Tpartition_ptr->Tnodelist_root = node_ptr;
	}

	if (strcmp(Stored[V_SSL], SSLOptions[0]) == 0)
		Tpartition_ptr->C_ssl = 0;
	else
		Tpartition_ptr->C_ssl = 1;
	strcpy(Tpartition_ptr->SZ_sslcertificate, Stored[V_SSLCERTIFICATE]);

	return I_VSIndex;
}

int menu_misc(void)
{
	int ch;
	
	Root->Help(EditHelp);
	
	Window Main ( 2,2,76,14 );
	Main.Border (1);
	Main.Title (WinTitle[9]);
	Window MainBox ( 4, 3, 72, 12 );

	Window Help ( 2,16,76,6);
	Help.Border (1);
	Help.Title (HelpTitle[0]);
	Window HelpBox ( 4,17,72,3 );

	IPField nfs_srv(&MainBox,1,26,40,NFSSERV);
	PathField nfs_bin(&MainBox,2,26,60,NFSBIN);

	SelectField interval(&MainBox,4,26,40,IntervalOptions,INTERVAL);
	SelectField loadbal(&MainBox,5,26,40,LoadBalancerOptions,LOADBAL);

	EmailField email(&MainBox,7,26,40,SERVERADMIN);
	email.SetBlankAllowed(true);

	SelectField zone(&MainBox,9,26,40,ZoneOptions,TIMEZONE);
	TimeField t(&MainBox,10,26,TIMEDATE);

	Field *MainFormItems[] =
	{
		&nfs_srv, &nfs_bin, &interval, &loadbal, &email, &zone, &t, NULL
	};
	Form MainForm(MainFormItems, &HelpBox);
	Top = &MainForm;
	MainForm.Render();

	do {
		ch = getch();
		if(!t.changed) {
			Field *cur = MainForm.items[MainForm.selected];
			if(cur != &t) {
				t.Display(0);
				MainBox.Update();
			}
			cur->Display(!cur->changed);
			cur->win->Update();
		}
		ch = MainForm.Edit(ch);
	} while(ch != KEY_F(10));

	// Save the integer interval for card services
	Tconfigp_ptr->UI_cardsvcsscanfreq = atoi(strtok(Stored[INTERVAL]," "));
	
	return 0;
}

int ConfirmExit()
{
	int ch, ret = 0;
	Window Confirm(10,8,60,8);
	Confirm.Border(1);
	Confirm.Title( "Confirm save operation" );
	Confirm.Print(4,2,"About to save your configuration data to disk.");
	Confirm.Print(4,4,"Press the 'Y' key to continue and save your changes.");
	Confirm.Print(4,5,"Press the 'N' key to abandon your changes and exit.");
	Confirm.Update();
	do ch = tolower( wgetch( Confirm.win ) );
	while ( ch != 'y' && ch != 'n' );
	
	if ( ch == 'n' )
	{
		Window Splash( 6,8,68,9 );
		Splash.Border(1);
		Splash.Print( 4,2,"Configuration process aborted.  No changes have been made");
		Splash.Print( 4,3,"to your Zaphod configuration file.  If you wish to re-run");
		Splash.Print( 4,4,"this configuration program, just type zaphodcfg again.");
		Splash.Print( 4,6,"Press any key to continue...");
		Splash.Update();
		while ( getch() == ERR );
	}
	else if ( (ret = config_save(Tconfigp_ptr)) < 0 )
	{
		Window Splash( 6,8,68,9 );
		Splash.Border(1);
		Splash.Print( 4,2,"An error occured while trying to save your changes." );
		Splash.Print( 4,3,"Please ensure you have permission to modify the Zaphod");
		Splash.Print( 4,4,"configuration files and re-run zaphodcfg.");
		Splash.Print( 4,6,"Press any key to continue...");
		Splash.Update();
		while ( getch() == ERR );
		ret  = -1;
	}
	else
	{
		// Signal card services that changes have been made
		// NOTE: we will eventually need to distinguish system
		// and apache changes as the former requires a system
		// restart. The splash message will have to be changed
		// accordingly.
		int pid;
		pid = getCardServicesPID();
		if (pid > 0)
		{
			// Tell the user that the settings have been saved and
			// that a restart may be necessary (if system settings
			// changed)
			Window Splash( 6,8,68,9 );
			Splash.Border(1);
			Splash.Print( 4,2,"Your changes have been saved.");
			if (ret != 0)
			{
			    if (ret == 1)
				Splash.Print( 4,3,"Zaphod will now restart in order for settings to be applied." );
			    else
				Splash.Print( 4,3,"Zaphod will now apply these settings." );
			}
			Splash.Print( 4,5,"Press any key to continue...");
			Splash.Update();

			while (getch() == ERR);
			
			// Send signal(s) SIGUSR1 or SIGUSR2 to
			// CardServices for system config changes and web
			// server config changes respectively.
			if (ret != 0)
			{
			    if (ret == 1)
				kill(pid, SIGUSR1);
			    else if (ret == 2)
				kill(pid, SIGUSR2);
			}
		}
		else
		{
			// Tell the user that changes will not take effect
			// until the system is restarted.
		}
		ret = 0;
	}
	destroy_partitions(Tconfigp_ptr);
	free(Tconfigp_ptr);
	return( ret );
}

bool ConfirmDelete(char* SZ_server)
{
	int	ch;
	Window	Confirm(5,7,70,8);
	char	buf[255];

	Confirm.Border(1);
	Confirm.Title( "Confirm Deletion" );
	sprintf(buf, "Are you sure you want to delete %s ?", SZ_server);
	Confirm.Print(4,2,buf);
	sprintf(buf, "Press the 'Y' key to delete %s", SZ_server);
	Confirm.Print(4,4,buf);
	Confirm.Print(4,5,"Press the 'N' key to cancel.");
	Confirm.Update();
	do
	{
		ch = tolower( wgetch( Confirm.win ) );
	}
	while ( ch != 'y' && ch != 'n' );
	
	return ( ch == 'y' );
}
